{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE NamedFieldPuns #-}

module System.Hardware.Joystick.Linux
  ( countJoysticks
  , openJoystick
  ) where

import Foreign hiding (void)

import Control.Monad (void)
import Data.List (genericSplitAt)
import Data.Maybe (isJust)
import Data.Bits ((.&.))
import Data.ByteString.Internal (unsafeCreate)
import Safe (readMay)
import System.Directory (getDirectoryContents)
import System.Posix.Types (Fd(..))
import System.Posix.IO.ByteString
         ( openFd
         , closeFd
         , defaultFileFlags
         , OpenFileFlags(..)
         , OpenMode(..) )

import qualified Data.ByteString as BS
import qualified Data.ByteString.Char8 as BS8

import System.Hardware.Joystick.Internal
import System.Hardware.Joystick.Types
import System.Hardware.Joystick.Linux.FFI

data Js = Js
  { jsFd      :: !Fd
  , jsVersion :: !Int
  , jsButtons :: !Int
  , jsAxes    :: !Int
  , jsName    :: !ByteString }
  deriving (Show, Eq)


paths :: IO [String]
paths =
  do c <- getDirectoryContents dir
     let js = filter forJs c
         ps = fmap (dir ++) js
     return ps
 where
  dir = "/dev/input/"
  forJs str =
    let (a, b) = genericSplitAt 2 str
    in a == "js" && isJust (readMay b :: Maybe Int)


-- | The number of available joysticks.
countJoysticks :: IO Int
countJoysticks =
  do ps <- paths
     return (length ps)


-- | Open a joystick by its index, starting from 0.
openJoystick :: Int -> IO Joystick
openJoystick n =
  do js <- openJs n
     return $ Joystick
       { joystickVersion = jsVersion js
       , joystickButtons = jsButtons js
       , joystickAxes    = jsAxes js
       , joystickName    = jsName js
       , joystickPoll    = joystickPollMethod js
       , joystickClose   = closeFd (jsFd js) }

openJs :: Int -> IO Js
openJs n =
  do ps <- paths 
     let p = BS8.pack $ ps !! n
     fd@(Fd c_fd) <- open p
     vers <- c_js_version c_fd
     btns <- c_js_buttons c_fd
     axes <- c_js_axes c_fd
     let name = BS.takeWhile (/= 0) (get_name c_fd)
     return $ Js
       { jsFd      = fd
       , jsVersion = fromIntegral vers
       , jsButtons = fromIntegral btns
       , jsAxes    = fromIntegral axes
       , jsName    = name }
 where
  open p = openFd p ReadOnly Nothing defaultFileFlags{ nonBlock = True }
  get_name fd = unsafeCreate 256
                             (\ptr -> void $ c_js_name fd ptr 256)

joystickPollMethod :: Js -> IO [JoystickEvent]
joystickPollMethod js@Js{ jsFd = (Fd fd), .. } =
  do ptr <- new $ JsEvent 0 0 0 0
     loop ptr []
 where
  loop p es =
    do r <- c_js_event_read fd p
       evt <- peek p
       if r == 1
          then case toJoystickEvent evt of
                 Nothing -> loop p es
                 Just e -> loop p (e:es)
          else do free p
                  return es

toJoystickEvent :: JsEvent -> Maybe JoystickEvent
toJoystickEvent JsEvent{..} =
  do etype <- toJsEventType jsEventType
     return $ case etype of
       JsEventButton -> JoystickEventButton
         { joystickEventButtonIndex   = fromIntegral jsEventNumber
         , joystickEventButtonPressed = jsEventValue == 1 }
       JsEventAxis -> JoystickEventAxis
         { joystickEventAxisIndex = fromIntegral jsEventNumber
         , joystickEventAxisValue = normalize jsEventValue }
 where
  normalize n =
    let d = if n > 0 then maxBound else minBound :: Int16
    in (fromIntegral n) / (abs $ fromIntegral d)

